require("main.constants")

local DEFAULT_NEAR = 0.1
local DEFAULT_FAR = 1000
local DEFAULT_FOV = 45
local DEFAULT_DISTANCE = 10
local DEFAULT_HEADING = 45

local camera_update_frustum = function(camera)
    camera.frustum.frustum = camera.projection * camera.view
end

local camera_update_window = function(camera, window_width, window_height)
    local w = math.max(1, window_width)
    local h = math.max(1, window_height)

    camera.aspect = w / h
    camera.projection = vmath.matrix4_perspective(math.rad(camera.fov), camera.aspect, camera.near, camera.far)

    camera_update_frustum(camera)
end

local camera_normalize_heading = function(camera)
    while camera.heading > 360 do
        camera.heading = camera.heading - 360
    end
    while camera.heading < 0 do
        camera.heading = camera.heading + 360
    end
end

local camera_update_heading = function(camera)
    local heading_rad = camera.heading * DEGREES
    local dx = math.cos(heading_rad)
    local dz = math.sin(heading_rad)

    camera.position.x = camera.target.x + (dx * camera.distance)
    camera.position.y = camera.target.y + camera.distance
    camera.position.z = camera.target.z + (dz * camera.distance)
    camera.view       = vmath.matrix4_look_at(camera.position, camera.target, camera.up)
end

local create = function(window_width, window_height, near, far, fov, distance, heading)
    local camera = 
    {
        near        = near == nil and DEFAULT_NEAR or near,
        far         = far == nil and DEFAULT_FAR or far,
        fov         = fov == nil and DEFAULT_FOV or fov,
        distance    = distance == nil and DEFAULT_DISTANCE or distance,
        heading     = heading == nil and DEFAULT_HEADING or heading,
        position    = vmath.vector3(0, 0, 0),
        target      = vmath.vector3(0, 0, 0),
        --right       = vmath.vector3(1, 0, 0),
        up          = vmath.vector3(0, 1, 0),
        frustum     = {},
    }

    camera_update_heading(camera)
    camera_update_window(camera, window_width, window_height)

    return camera
end

local update_heading = function(camera, step, dt)
    camera.heading = camera.heading + (step * dt)

    camera_normalize_heading(camera)
    camera_update_heading(camera)
    camera_update_frustum(camera)
end

local set_target = function(camera, x, y, z)
    camera.target.x = x
    camera.target.y = y
    camera.target.z = z

    camera_update_heading(camera)
    camera_update_frustum(camera)
end

return 
{
    create = create,
    update_window = camera_update_window,
    update_heading = update_heading,
    set_target = set_target,
}
